import { Discord } from "./discord";
import { Glue } from "./lib/glue";
import { Matrix } from "./lib/matrix";
import { prisma } from "./lib/prisma";
import {
  IClientEvent,
  MatrixEphemeralEvent,
  MatrixEphemeralTyping,
  MatrixError,
  MatrixRoomMembership,
  MatrixRoomMessage,
  MatrixRoomRedaction,
  MatrixStickerEvent,
} from "./types/matrix";

/**
 * Homeserver friendly name (not direct access)
 */
const MATRIX_HOMESERVER =
  process.env.MATRIX_HOMESERVER_HOST ?? process.env.MATRIX_HOMESERVER;

class MatrixHandler_ {
  private handledTransactionIDs: string[] = [];

  async handleTransaction(
    txnId: string,
    events: IClientEvent[],
    ephemeral: MatrixEphemeralEvent[]
  ) {
    if (this.handledTransactionIDs.indexOf(txnId) > -1) return;

    let matrixClient;
    try {
      matrixClient = await Matrix.for("_discord_bot");
    } catch (e) {
      console.error("Failed to get main bridge account", e);
      return;
    }

    for (const event of ephemeral) {
      console.log("[e:Matrix->]", event);

      switch (event.type) {
        case "m.typing": {
          await this.handleTyping(event);
          break;
        }
      }
    }

    for (const event of events) {
      console.log("[Matrix->]", event);

      if (process.env.VOID_ALL_TRANSACTIONS) {
        console.log("Voided transaction ^^");
        continue;
      }

      switch (event.type) {
        case "m.room.member": {
          const event2: MatrixRoomMembership = event as any;

          await this.handleRoomMembership(event2);
          break;
        }
        case "m.room.redaction": {
          const event2: MatrixRoomRedaction = event as any;

          await this.handleDeleteEvent(event2);
          break;
        }
        case "m.room.message": {
          if (event.sender.startsWith("@_discord")) continue;
          const event2: MatrixRoomMessage = event as any;

          if (
            "m.relates_to" in event2.content &&
            "rel_type" in event2.content["m.relates_to"] &&
            event2.content["m.relates_to"].rel_type === "m.replace"
          ) {
            // message edit
            await this.handleEditMessage(event2);
          } else {
            // this is a new message
            await this.handleNewMessage(event2);
          }
          break;
        }
        case "m.sticker": {
          if (event.sender.startsWith("@_discord")) continue;
          const event2: MatrixStickerEvent = event as any;

          // these messages cannot be edited
          await this.handleSticker(event2);
          break;
        }
        default: {
          console.warn("Unhandled event (unknown type)", event);

          await matrixClient.replyNotice(
            event as any,
            "**Can't bridge** Unknown event type: " + event.type
          );
        }
      }
    }
  }

  /**
   * Handle user query from the homeserver
   *
   * Ensure account is available before returning
   *
   * @param userId
   * @returns
   */
  async handleUserQuery(
    userId: string
  ): Promise<{ status: 200 } | { status: 401 | 404; data: MatrixError }> {
    return { status: 200 };
  }

  /**
   * handle room query from the homeserver
   *
   * Must create the room before returning
   * Room data (like topic) can be set before returning
   *
   * @param alias
   */
  async handleRoomQuery(
    alias: string
  ): Promise<{ status: 200 } | { status: 401 | 404; data: MatrixError }> {
    return { status: 200 };
  }

  async handleSticker(event: MatrixStickerEvent) {
    const bridges = await Glue.getBridges({ matrix_room: event.room_id });

    const profile = await Matrix.getUserProfile(event.sender);

    for (const bridge of bridges) {
      try {
        const discMsg = await Discord.sendMessage(
          bridge.discord_guild_id,
          bridge.discord_channel_id,
          profile,
          "",
          {
            attachment: {
              url: Matrix.getMediaURL(event.content.url, "internal") as string,
              name:
                "sticker." +
                (event.content.info.mimetype?.split("/").at(-1) || "png"),
            },
          }
        );

        await prisma.matrixMessages.create({
          data: {
            matrix_event_id: event.event_id,
            matrix_room_id: event.room_id,
            matrix_author_id: event.sender,
            discord_id: discMsg.id,
            discord_channel_id: discMsg.channelId,
            discord_guild_id: discMsg.guildId!,
          },
        });
      } catch (e) {
        console.error("Failed to bridge message", e);
      }
    }
  }

  /**
   * Handle m.typing messages
   * @param event
   * @returns
   */
  async handleTyping(event: MatrixEphemeralTyping) {
    const user_ids = event.content.user_ids.filter(
      (uid) => !uid.startsWith("@_discord_")
    );
    if (user_ids.length === 0) return; // discord doesn't have a "stop typing event"

    const bridges = await Glue.getBridges({ matrix_room: event.room_id });
    if (!bridges) return;

    for (const bridge of bridges) {
      const { guild, channel } = await Discord.getGuildChannel(
        bridge.discord_guild_id,
        bridge.discord_channel_id
      );

      channel.sendTyping();
    }
  }

  async handleRoomMembership(event: MatrixRoomMembership) {
    if (event.content.membership === "invite") {
      // if someone is being invited, we only want to react when it's related to one of the virual accounts

      if (event.state_key === "@_discord_bot:" + MATRIX_HOMESERVER) {
        const matrixClient = await Matrix.for("_discord_bot");

        if (event.sender !== process.env.MATRIX_ADMIN) {
          // only the admin account should be able to add the bot
          await matrixClient.leaveRoom(event.room_id);
          return;
        }

        // auto accept if it's the main bot that's being invited

        await matrixClient.joinRoom(event.room_id);
      } else {
        // for another bot account
        // we only want the main bot to be able to invite virtual people
        if (event.sender !== `@_discord_bot:${MATRIX_HOMESERVER}`) return;

        const user = await Matrix.for(
          event
            .state_key!.replace("@", "")
            .replace(":" + MATRIX_HOMESERVER, "") as any
        );
        await user.joinRoom(event.room_id);
      }
    }

    if (event.content.membership === "join") {
      if (event.prev_content?.membership === "join") return; // not a membership join
      if (event.state_key?.startsWith("@_discord_")) return;
      await this.handleRoomJoin(event);
    }

    if (event.content.membership === "leave") {
      if (event.prev_content?.membership === "leave") return; // not a membership leave

      if (event.state_key && !event.state_key.startsWith("@_discord_"))
        await this.handleRoomLeave(event);
    }

    if (event.content.membership === "ban") {
      // handle banning
    }
  }

  async handleRoomJoin(event: MatrixRoomMembership) {
    // const linkedDiscord = await this.getDiscordFromAlias(event.room_id);
    // if (!linkedDiscord) return;
    // const profile = await Matrix.getUserProfile(event.sender);
    // await Discord.sendMessage(
    //   linkedDiscord.guildId,
    //   linkedDiscord.channelId,
    //   profile,
    //   ":arrow_right: *joined room*"
    // );
  }

  async handleRoomLeave(event: MatrixRoomMembership) {
    // const linkedDiscord = await this.getDiscordFromAlias(event.room_id);
    // if (!linkedDiscord) return;
    // const profile = await Matrix.getUserProfile(event.sender);
    // await Discord.sendMessage(
    //   linkedDiscord.guildId,
    //   linkedDiscord.channelId,
    //   profile,
    //   ":arrow_left: *left room*"
    // );
  }

  async handleNewMessage(message: MatrixRoomMessage) {
    const bridged = await Glue.getBridges({
      matrix_room: message.room_id,
    });
    if (!bridged.length) return;

    let matrixClient;
    try {
      matrixClient = await Matrix.for("_discord_bot");
    } catch (e) {
      console.error("Failed to get main bot account", e);
      return;
    }

    const inReplyTo_eventID =
      "m.relates_to" in message.content &&
      "m.in_reply_to" in message.content["m.relates_to"] &&
      message.content["m.relates_to"]["m.in_reply_to"]?.event_id;

    /**
     * If incoming message is in reply to another message, we need to find the possible bridged message
     *
     * Steps:
     * 1. Check that incoming message is "in reply to"
     * 2. Look for (matrix -> discord) message -> return discord id
     * 3. Look for (discord -> matrix) message -> return discord id
     */
    let inReplyTo: NonNullable<
      Parameters<(typeof Discord)["sendMessage"]>[4]
    >["replyTo"] = undefined;

    if (inReplyTo_eventID) {
      const matrixMessage = await prisma.matrixMessages.findFirst({
        where: { matrix_event_id: inReplyTo_eventID },
      });

      if (matrixMessage) {
        inReplyTo = {
          messageId: matrixMessage.discord_id,
          who: matrixMessage.matrix_author_id,
        };
      } else {
        const discordMessage = await prisma.discordMessages.findFirst({
          where: { matrix_event_id: inReplyTo_eventID },
        });

        if (discordMessage) {
          inReplyTo = {
            messageId: discordMessage.discord_id,
            who: discordMessage.discord_author_id,
          };
        }
      }
    }

    const profile = await Matrix.getUserProfile(message.sender);

    switch (message.content.msgtype) {
      case "m.text": {
        for (const room of bridged) {
          try {
            const discMsg = await Discord.sendMessage(
              room.discord_guild_id,
              room.discord_channel_id,
              profile,
              Discord.matrixToDiscord(message),
              {
                replyTo: inReplyTo,
              }
            );

            await prisma.matrixMessages.create({
              data: {
                matrix_event_id: message.event_id,
                matrix_room_id: message.room_id,
                matrix_author_id: message.sender,
                discord_id: discMsg.id,
                discord_channel_id: discMsg.channelId,
                discord_guild_id: discMsg.guildId!,
              },
            });
          } catch (e) {
            console.error("Failed to bridge to room", room, e);
          }
        }
        break;
      }
      case "m.file":
      case "m.video":
      case "m.audio":
      case "m.image": {
        for (const room of bridged) {
          try {
            const discMsg = await Discord.sendMessage(
              room.discord_guild_id,
              room.discord_channel_id,
              profile,
              Discord.matrixToDiscord(message),
              {
                replyTo: inReplyTo,
                attachment: {
                  url: Matrix.getMediaURL(
                    message.content.url as any,
                    "internal"
                  ) as string,
                  name: message.content.body,
                },
              }
            );

            await prisma.matrixMessages.create({
              data: {
                matrix_event_id: message.event_id,
                matrix_room_id: message.room_id,
                matrix_author_id: message.sender,
                discord_id: discMsg.id,
                discord_channel_id: discMsg.channelId,
                discord_guild_id: discMsg.guildId!,
              },
            });
          } catch (e) {
            console.error("Failed to bridge to room", room, e);
          }
        }
        break;
      }
      default: {
        console.warn("Unhandled new room message type", message);

        await matrixClient.replyNotice(
          message,
          "**Can't bridge** Unknown message type: " + message.content.msgtype
        );
      }
    }
  }

  async handleEditMessage(message: MatrixRoomMessage) {
    if (
      !("m.relates_to" in message.content) ||
      !("event_id" in message.content["m.relates_to"]) ||
      !("m.new_content" in message.content)
    )
      return;

    switch (message.content.msgtype) {
      case "m.text": {
        const db_messages = await prisma.matrixMessages.findMany({
          where: {
            matrix_event_id: message.content["m.relates_to"]!.event_id,
          },
        });

        for (const db_message of db_messages) {
          try {
            const discEdit = await Discord.editMessage(
              db_message.discord_guild_id,
              db_message.discord_channel_id,
              db_message.discord_id,
              message.content["m.new_content"]!.body
            );
          } catch (e) {
            console.error("Edit message discord error", db_message, e);
          }
        }
        break;
      }
      default:
        console.warn("Unhandled edit room message type", message);
    }
  }

  async handleDeleteEvent(event: MatrixRoomRedaction) {
    const db_matrix = await prisma.matrixMessages.findMany({
      where: {
        matrix_event_id: event.content.redacts,
      },
    });

    // clean matrix -> discord message
    for (const db_message of db_matrix) {
      try {
        // delete message on the discord side
        await Discord.deleteMessage(
          db_message.discord_guild_id,
          db_message.discord_channel_id,
          db_message.discord_id
        );
      } catch (e) {
        console.error("Failed to delete message", db_message, e);
      }
    }

    const db_discord = await prisma.discordMessages.findMany({
      where: {
        matrix_event_id: event.content.redacts,
      },
    });

    // clean discord -> matrix message
    for (const db_message of db_discord) {
      try {
        await prisma.discordMessages.delete({
          where: {
            id: db_message.id,
          },
        });
        // delete message on the discord side
        await Discord.deleteMessage(
          db_message.discord_guild_id,
          db_message.discord_channel_id,
          db_message.discord_id,
          false
        );
      } catch (e) {
        console.error("Failed to delete message", db_message, e);
      }
    }
  }
}

export const MatrixHandler = new MatrixHandler_();
